import { promisify } from 'util';
import Module from 'module';
import { writeFile, mkdir, unlink, symlink } from 'fs/promises';
import { resolve, relative, dirname, join, basename } from 'upath';
import chalk from 'chalk';
import consola from 'consola';
import rimraf from 'rimraf';
import defu from 'defu';
import prettyBytes from 'pretty-bytes';
import jiti from 'jiti';
import { rollup } from 'rollup';
import commonjs from '@rollup/plugin-commonjs';
import nodeResolve from '@rollup/plugin-node-resolve';
import alias from '@rollup/plugin-alias';
import esbuild from 'rollup-plugin-esbuild';
import dts from 'rollup-plugin-dts';
import { resolveSchema, generateMarkdown, generateTypes } from 'untyped';
import { pascalCase } from 'scule';
import { mkdist } from 'mkdist';

function validateDependencies(ctx) {
  const usedDependencies = new Set();
  const unusedDependencies = new Set(Object.keys(ctx.pkg.dependencies || {}));
  const implicitDependnecies = new Set();
  for (const id of ctx.usedImports) {
    unusedDependencies.delete(id);
    usedDependencies.add(id);
  }
  if (Array.isArray(ctx.dependencies)) {
    for (const id of ctx.dependencies) {
      unusedDependencies.delete(id);
    }
  }
  for (const id of usedDependencies) {
    if (!ctx.externals.includes(id) && !id.startsWith("chunks/") && !ctx.externals.includes(id.split("/")[0])) {
      implicitDependnecies.add(id);
    }
  }
  if (unusedDependencies.size) {
    consola.warn("Potential unused dependencies found:", Array.from(unusedDependencies).map((id) => chalk.cyan(id)).join(", "));
  }
  if (implicitDependnecies.size) {
    consola.warn("Potential implicit dependencies found:", Array.from(implicitDependnecies).map((id) => chalk.cyan(id)).join(", "));
  }
}

async function rollupBuild(ctx) {
  if (ctx.stub) {
    for (const entry of ctx.entries.filter((entry2) => entry2.builder === "rollup")) {
      const input = resolve(ctx.rootDir, entry.input);
      const output = resolve(ctx.rootDir, ctx.outDir, entry.name);
      await writeFile(output + ".js", `module.exports = require('jiti')()('${input}')`);
      await writeFile(output + ".mjs", `export * from '${input}'`);
      await writeFile(output + ".d.ts", `export * from '${input}'`);
    }
    return;
  }
  const rollupOptions = getRollupOptions(ctx);
  if (!Object.keys(rollupOptions.input).length) {
    return;
  }
  const buildResult = await rollup(rollupOptions);
  const allOutputOptions = rollupOptions.output;
  for (const outputOptions of allOutputOptions) {
    const {output} = await buildResult.write(outputOptions);
    for (const entry of output.filter((e) => e.type === "chunk")) {
      for (const id of entry.imports) {
        ctx.usedImports.add(id);
      }
      if (entry.isEntry) {
        ctx.buildEntries.push({
          path: relative(ctx.rootDir, resolve(outputOptions.dir, entry.fileName)),
          bytes: entry.code.length * 4,
          exports: entry.exports
        });
      }
    }
  }
  rollupOptions.plugins = rollupOptions.plugins || [];
  rollupOptions.plugins.push(dts());
  const typesBuild = await rollup(rollupOptions);
  await typesBuild.write({
    dir: resolve(ctx.rootDir, ctx.outDir),
    format: "esm"
  });
}
function getRollupOptions(ctx) {
  const extensions = [".ts", ".mjs", ".js", ".json"];
  return {
    context: ctx.rootDir,
    input: Object.fromEntries(ctx.entries.filter((entry) => entry.builder === "rollup").map((entry) => [entry.name, resolve(ctx.rootDir, entry.input)])),
    output: [
      {
        dir: resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].js",
        chunkFileNames: "chunks/[name].js",
        format: "cjs",
        exports: "auto",
        preferConst: true
      },
      {
        dir: resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].mjs",
        chunkFileNames: "chunks/[name].mjs",
        format: "esm",
        exports: "auto",
        preferConst: true
      }
    ],
    external(id) {
      if (id[0] === "." || id[0] === "/" || id.includes("src/")) {
        return false;
      }
      const isExplicitExternal = !!ctx.externals.find((ext) => id.includes(ext));
      if (!isExplicitExternal) {
        consola.warn(`Inlining external ${id}`);
      }
      return isExplicitExternal;
    },
    onwarn(warning, rollupWarn) {
      if (!warning.code || !["CIRCULAR_DEPENDENCY"].includes(warning.code)) {
        rollupWarn(warning);
      }
    },
    plugins: [
      alias({}),
      nodeResolve({
        extensions
      }),
      esbuild({
        target: "node12",
        loaders: {
          ".json": "json"
        }
      }),
      commonjs({
        extensions
      })
    ]
  };
}

async function typesBuild(ctx) {
  for (const entry of ctx.entries.filter((entry2) => entry2.builder === "untyped")) {
    const srcConfig = await import(resolve(ctx.rootDir, entry.input)).then((r) => r.default);
    const genDir = resolve(ctx.rootDir, ctx.genDir, entry.name);
    const defaults = entry.defaults || {};
    const schema = resolveSchema(srcConfig, defaults);
    await mkdir(dirname(genDir)).catch(() => {
    });
    await mkdir(genDir).catch(() => {
    });
    await writeFile(resolve(genDir, `${entry.name}.md`), generateMarkdown(schema));
    await writeFile(resolve(genDir, `${entry.name}.schema.json`), JSON.stringify(schema, null, 2));
    await writeFile(resolve(genDir, `${entry.name}.defaults.json`), JSON.stringify(defaults, null, 2));
    await writeFile(resolve(genDir, `${entry.name}.d.ts`), "export " + generateTypes(schema, pascalCase(entry.name + "-schema")));
  }
}

async function mkdistBuild(ctx) {
  for (const entry of ctx.entries.filter((e) => e.builder === "mkdist")) {
    if (ctx.stub) {
      const srcDir = resolve(ctx.rootDir, entry.input);
      const outDir = resolve(ctx.rootDir, ctx.outDir, entry.name);
      await unlink(outDir).catch(() => {
      });
      await symlink(srcDir, outDir);
    } else {
      const {writtenFiles} = await mkdist({
        rootDir: ctx.rootDir,
        srcDir: entry.input,
        distDir: join(ctx.outDir, entry.name),
        format: entry.format,
        cleanDist: false
      });
      ctx.buildEntries.push({
        path: join(ctx.outDir, entry.name),
        chunks: [`${writtenFiles.length} files`]
      });
    }
  }
}

async function build(rootDir, stub) {
  rootDir = resolve(process.cwd(), rootDir || ".");
  const _require = jiti(rootDir);
  let buildConfigFile;
  try {
    buildConfigFile = _require.resolve("./build.config");
  } catch (e) {
  }
  const buildConfig = buildConfigFile ? _require("./build.config").default : {};
  const pkg = _require("./package.json");
  const ctx = defu(buildConfig, {
    pkg,
    rootDir,
    entries: [],
    dependencies: [],
    externals: [...Module.builtinModules],
    srcDir: "src",
    outDir: "dist",
    genDir: ".gen",
    untyped: void 0,
    clean: true,
    stub,
    buildEntries: [],
    usedImports: new Set()
  });
  ctx.entries = ctx.entries.map((entry) => typeof entry === "string" ? {input: entry} : entry);
  for (const entry of ctx.entries) {
    if (typeof entry.name !== "string") {
      entry.name = basename(entry.input);
    }
    if (!entry.input) {
      throw new Error("Missing entry input: " + dumpObject(entry));
    }
    if (!entry.builder) {
      entry.builder = entry.input.endsWith("/") ? "mkdist" : "rollup";
    }
  }
  if (pkg.dependencies) {
    ctx.externals.push(...Object.keys(pkg.dependencies));
  }
  consola.info(chalk.cyan(`Building ${pkg.name}`));
  if (process.env.DEBUG) {
    consola.info(`${chalk.bold("Root dir:")} ${ctx.rootDir}
  ${chalk.bold("Entries:")}
  ${ctx.entries.map((entry) => "  " + dumpObject(entry)).join("\n  ")}
`);
  }
  if (ctx.clean) {
    const outDir = resolve(ctx.rootDir, ctx.outDir);
    await unlink(outDir).catch(() => {
    });
    await promisify(rimraf)(outDir);
    await mkdir(outDir).catch(() => {
    });
  }
  await typesBuild(ctx);
  await mkdistBuild(ctx);
  await rollupBuild(ctx);
  consola.success(chalk.green("Build succeed for " + pkg.name));
  for (const entry of ctx.buildEntries) {
    consola.log(`  ${chalk.bold(entry.path)} (` + [
      entry.bytes && `size: ${chalk.cyan(prettyBytes(entry.bytes))}`,
      entry.exports && `exports: ${chalk.gray(entry.exports.join(", "))}`,
      entry.chunks && `chunks: ${chalk.gray(entry.chunks.join(", "))}`
    ].filter(Boolean).join(", ") + ")");
  }
  validateDependencies(ctx);
  consola.log("");
}
function dumpObject(obj) {
  return "{ " + Object.keys(obj).map((key) => `${key}: ${JSON.stringify(obj[key])}`).join(", ") + " }";
}

export { build as b };
